#ifndef lint
static char *RCSid = "$Id: stack.c,v 1.1 1999/08/31 20:32:32 mark Exp $";
#endif

/*
 *  The Regina Rexx Interpreter
 *  Copyright (C) 1992-1994  Anders Christensen <anders@pvv.unit.no>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Consept of implementation:

 * In this description, the stack is described as growing upwards

 * The stack has two pointers to the ends of a double-linked list,
 * which contains entries consisting of just a pointer to the text (in
 * a struct streng) in addition to pointers to next and previous
 * entry.  Last box' next, and first box' prev is both NIL. The text
 * pointer points to the text of that line in the stack.

 * If the text pointer is NIL, then that box in the stack-structure is
 * just symbolizing a stack buffer mark. The integer variable
 * 'buffers' contains the current number of buffermarks in the stack.
 * A buffer is the lines between two buffermarks. The first buffer mark
 * (for buffer 0) is not marked, but is implicit the bottom of the
 * stack. The last buffer is the lines between the uppermost
 * buffermark and the top of the stack.

 * Initialy all lines are put in buffer 0. When a new buffer is
 * created, lines pushed lifo is put on top of that buffer, and lines
 * queued fifo are put in the bottom of that buffer (i.e. not in the
 * bottom of buffer 0.) Lines pushed on the stack are 'eaten' and
 * there must not be any referances to them in other parts of the
 * program after they have been pushed.

 * When reading lines from the stack, and all lines in the current
 * buffer has been read, the buffer-mark will be removed, and lines
 * are read from the buffer underneath. Lines read from the stack must
 * be free'ed by the routine that popped them from the stack.

 * When the whole stack is empty, lines are read from standard input.

 * Buffer 0 is contains the lines between the bottom of the stack, and
 * the first buffermark on the stack, the first buffer, is the lines
 * between buffermark 1 and buffer mark 2.

 * When creating a buffer, the value returned will be the number of
 * buffermarks in the stack after the new buffermark is created.
 * When destroying buffers, the parameter given will equal the number
 * of the lowest buffermark to be destroyed (i.e dropbuf 4) will leave
 * 3 buffermarks on the stack).

 * Possible extentions:
 *  o A new 'hard-buffer', that is not removed when a line is popped
 *    from it while it is empty
 */

#include "rexx.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

/*
 * Hide the definition of the stack from the rest of the program
 */
typedef struct stacklinestruct *stacklineptr ;

typedef struct stacklinestruct
{
   stacklineptr next, prev ;
   streng *contents ;
} stackline ;

typedef struct { /* stk_tsd: static variables of this module (thread-safe) */
   /*
    * Pointers to first and last line in the stack.
    */
   stacklineptr lastline0 ;
   stacklineptr firstline0 ;
   /*
    * The number of buffermarks in the stack
    */
   int          buffers ;
   /*
    * Pointers to first and last entry in the temporary stack
    */
   stacklineptr firstbox ;
   stacklineptr lastbox ;
} stk_tsd_t; /* thread-specific but only needed by this module. see
              * init_stacks
              */

/* init_stacks initializes the module.
 * Currently, we set up the thread specific data.
 * The function returns 1 on success, 0 if memory is short.
 */
int init_stacks( tsd_t *TSD )
{
   stk_tsd_t *st;

   if (TSD->stk_tsd != NULL)
      return(1);

   if ((st = TSD->stk_tsd = MallocTSD(sizeof(stk_tsd_t))) == NULL)
      return(0);
   memset(st,0,sizeof(stk_tsd_t));  /* correct for all values */
   return(1);
}
/*
 * This routines accumulates stackentries, but without pushing them
 * on the actual stack, used when stacking the output of an external
 * command. If such a command also reads from the stack, the output
 * can not be flushed to the stack before the command has finished
 * reading (else, the command might read its own output.)
 *
 * Parameter is the streng to stack.
 */
void tmp_stack( const tsd_t *TSD, streng *value, int is_fifo )
{
   stacklineptr ptr=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   assert( ((st->lastbox)&&(st->firstbox)) || ((!st->lastbox)&&(!st->firstbox)) ) ;
   assert( (value) && (Str_max(value) >= Str_len(value)) ) ;

#if OLD_OPTIONS
   if (TSD->currlevel->u.options.flushstack)
#else
   if ( get_options_flag( TSD->currlevel, EXT_FLUSHSTACK ) )
#endif
   {
      if (is_fifo)
         stack_fifo( TSD, value ) ;
      else
         stack_lifo( TSD, value ) ;
   }
   else
   {
      ptr = MallocTSD( sizeof(struct stacklinestruct) ) ;
      ptr->contents = value ;
      ptr->next = NULL ;
      ptr->prev = st->lastbox ;


      if (st->firstbox)
         st->lastbox->next = ptr ;
      else
         st->firstbox = ptr ;
      st->lastbox = ptr ;
   }
}


/*
 * Flushes the content of the temporary stack created by (possibly
 * multiple) calls to tmp_stack().
 *
 * If parameter is true, lines are stacked fifo, which requires their
 * order to be reversed first.
 */
void flush_stack( const tsd_t *TSD, int is_fifo )
{
   stacklineptr ptr=NULL, tptr=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   if (st->firstbox==NULL)
   {
      /* nothing in temporary stack to flush */
      assert(st->lastbox==NULL) ;
      return ;
   }

   /* stack it either fifo or lifo */
   if (is_fifo)
   {
      /* if fifo, temporary stack must be reversed first */
      /* travelling in 'prev' direction, since stack is reversed */
      for (ptr=st->firstbox; ptr; ptr=ptr->prev )
      {
         tptr = ptr->prev ;
         ptr->prev = ptr->next ;
         ptr->next = tptr ;
      }

      /* temporary stack now in right order, link to top of real stack */
      st->firstbox->next = st->firstline0 ;
      if (st->firstline0)
         st->firstline0->prev = st->firstbox ;
      else
         st->lastline0 = st->firstbox ;
      st->firstline0 = st->lastbox ;
   }
   else
   {
      /* everything is in right order, just link them together */
      st->firstbox->prev = st->lastline0 ;
      if (st->lastline0)
         st->lastline0->next = st->firstbox ;
      else
         st->firstline0 = st->firstbox ;
      st->lastline0 = st->lastbox ;
   }

   /* reset the pointers, to signify that temporary stack is empty */
   st->lastbox = st->firstbox = NULL ;
}



/*
 * Pushes 'line' onto the REXX stack, lifo, and sets 'lastline' to
 *    point to the new line. The line is put on top of the current
 *    buffer.
 */
void stack_lifo( const tsd_t *TSD, streng *line )
{
   stacklineptr newbox=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   newbox = (stacklineptr)MallocTSD(sizeof(stackline)) ;
   if (st->lastline0)
   {
      st->lastline0->next = newbox ;
      newbox->prev = st->lastline0 ;
   }
   else
   {
      newbox->prev = NULL ;
      st->firstline0 = newbox ;
   }
   if (!line)
      st->buffers++ ;

   newbox->next = NULL ;
   newbox->contents = line ;
   st->lastline0 = newbox ;
}


/*
 * Puts 'line' on the REXX stack, fifo. This routine is similar to
 *    stack_lifo but the differences are big enough to have a separate
 *    routine. The line is put in the bottom of the current buffer,
 *    that is just above the uppermost buffer mark, or at the bottom
 *    of the stack, if there are no buffer marks.
 */
void stack_fifo( const tsd_t *TSD, streng *line )
{
   stacklineptr newbox=NULL, ptr=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   if (!line)
      st->buffers++ ;

   /* Bug: inserts into bottom of stack, not bottom of current buffer */
   newbox = (stacklineptr)MallocTSD(sizeof(stackline)) ;
   newbox->prev = newbox->next = NULL ;
   newbox->contents = line ;

   for (ptr=st->lastline0;(ptr)&&(ptr->contents);ptr=ptr->prev) ;

   if (ptr)
   {
      newbox->prev = ptr ;
      newbox->next = ptr->next ;
      if (ptr->next)
         ptr->next->prev = newbox ;
      else
         st->lastline0 = newbox ;
      ptr->next = newbox ;
   }
   else
   {
      newbox->next = st->firstline0 ;
      st->firstline0 = newbox ;
      if (newbox->next)
         newbox->next->prev = newbox ;
      if (!st->lastline0)
         st->lastline0 = newbox ;
   }
}


/*
 * Removes one (or several) buffers from the stack. The number of
 *    buffers to remove is decided by 'number', which is the number of
 *    buffers to remain on the stack. The global variable buffer
 *    contains the number of buffermarks currently in the stack.
 *
 * When number==2, buffer 0 (the implisit), and buffer 1 (the first
 *    buffer created by the user) are the only that remain. All lines
 *    from buffermark 2 (inclusive) and above are removed. Remember
 *    that buffer N is the lines following the N'th buffermark, until
 *    end-of-stack or another buffermark is reached.
 *
 * If number is less than zero, then abs(number) buffers is removed
 *    from the top of the stack (or until the stack is empty).

 * If the number is zero, the stack is emptied. If a buffer with a
 *    number that is higher than the current buffer is spesified,
 *    errorcode is retured.
 */

int drop_buffer( const tsd_t *TSD, int number )
{
   stacklineptr ptr=NULL, nptr=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   if (number<0)
      number = (st->buffers + number + 1) ;

   assert(st->buffers>=0) ;
   for (ptr=st->lastline0; (ptr) && (number<=st->buffers); ptr=nptr)
   {
      nptr = ptr->prev ;
      if (ptr->contents)
         Free_stringTSD(ptr->contents) ;
      else
         st->buffers-- ;
      FreeTSD(ptr) ;
      if (nptr)
         nptr->next = NULL ;
   }
   if ((st->lastline0=ptr) == NULL)
      st->firstline0 = NULL ;

   if (st->buffers<0)
      st->buffers = 0 ;

   return st->buffers ;
}


/*
 * Fetches a line from the top of the stack. If the current buffer
 *    does not contain any lines, it is removed and the second current
 *    buffer is search for a line etc. If there isn't any lines in the
 *    stack, a line is read from the standard input.
 */
streng *popline( tsd_t *TSD )
{
   streng *contents=NULL ;
   stacklineptr line=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   if ((line=st->lastline0) != NULL)
   {
      contents = line->contents ;
      st->lastline0 = line->prev ;
      if (!line->prev)
         st->firstline0 = NULL ;
      else
         line->prev->next = NULL ;

      FreeTSD(line) ;
      if (!contents)
      {
         st->buffers-- ;
         contents = popline( TSD ) ;
      }
   }
   else
   {
      int rc = HOOK_GO_ON ;
      if (TSD->systeminfo->hooks & HOOK_MASK(HOOK_PULL))
         rc = hookup_input( TSD, HOOK_PULL, &contents ) ;

      if (rc==HOOK_GO_ON)
         contents = readkbdline( TSD ) ;

      assert( contents ) ;
   }

   return contents ;
}


/*
 * Counts the lines in the stack.
 */
int lines_in_stack( const tsd_t *TSD )
{
   stacklineptr ptr=NULL ;
   int lines=0 ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   ptr = st->firstline0 ;
   for (lines=0;ptr;ptr=ptr->next)
      if (ptr->contents)
         lines++ ;

   return lines ;
}


/*
 * Returns boolean expression: is the stack empty?
 */
int stack_empty( const tsd_t *TSD )
{
   stacklineptr ptr=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   for (ptr=st->firstline0;ptr;ptr=ptr->next)
      if (ptr->contents)
         return 0 ;

   return 1 ;
}


#ifdef TRACEMEM
/*
 * Marks all chunks of dynamic allocated memory that are allocated
 *   to the stack subsystem.
 */
void mark_stack( const tsd_t *TSD )
{
   stacklineptr ptr=NULL ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   for (ptr=st->firstline0;ptr;ptr=ptr->next)
   {
      markmemory(ptr,TRC_STACKBOX) ;
      if (ptr->contents)
         markmemory(ptr->contents,TRC_STACKLINE) ;
   }
}
#endif


/*
 * Creates a new buffer, by putting a buffer mark at the top of the
 *    stack. (This could be implemented as {push_line(NULL)} and a
 *    special test for (line==NULL) in push_line())
 */
int make_buffer( const tsd_t *TSD )
{
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   stack_lifo( TSD, NULL ) ;
   return st->buffers ;
}



/*
 * Dumps the contents of the stack to standard error. Buffer marks are
 *    indicated in the printout.
 */
void type_buffer( const tsd_t *TSD )
{
   stacklineptr ptr=NULL ;
   char *cptr=NULL, *stop=NULL ;
   int counter=0 ;
   stk_tsd_t *st;

   st = TSD->stk_tsd;
   if (TSD->stddump == NULL)
      return;
   fprintf(TSD->stddump,"==> Lines: %d\n", lines_in_stack( TSD )) ;
   fprintf(TSD->stddump,"==> Buffer: %d\n", counter=st->buffers) ;
   for (ptr=st->lastline0; ptr; ptr=ptr->prev)
   {
      if (ptr->contents)
      {
         putc( '"', TSD->stddump ) ;
         stop = Str_end(ptr->contents) ;
         for (cptr=ptr->contents->value; cptr<stop; cptr++)
            putc( (isprint(*cptr) ? (*cptr) : '?'), TSD->stddump ) ;

         putc( '"', TSD->stddump ) ;
         putc( REGINA_EOL, TSD->stddump ) ;
      }
      else
         fprintf(TSD->stddump,"==> Buffer: %d\n",--counter) ; }

   fprintf(TSD->stddump,"==> End of Stack\n") ;
   fflush(TSD->stddump) ;
}
